import collections
import numpy as np
import itertools as it
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable


axesdict = collections.OrderedDict([("x", 0), ("y", 1), ("z", 2)])


def as_array(mesh, dim, val):
    val_array = np.empty(mesh.n + (dim,))
    if isinstance(val, (int, float)) and (dim == 1 or val == 0):
        val_array.fill(val)
    elif isinstance(val, (tuple, list, np.ndarray)) and len(val) == dim:
        val_array[..., :] = val
    elif isinstance(val, np.ndarray) and val.shape == val_array.shape:
        val_array = val
    elif callable(val):
        for index, point in zip(mesh.indices, mesh.coordinates):
            val_array[index] = val(point)
    else:
        raise TypeError("Unsupported type(val)={}.".format(type(val)))
    return val_array


def plane_info(*args, x=None, y=None, z=None):
    kwargs = locals()
    info = dict()
    if args:
        sliceaxis = args[0]
        info["point"] = None
    else:
        sliceaxis = [key for key in kwargs.keys()
                     if kwargs[key] is not None and key != "args"][0]
        info["point"] = kwargs[sliceaxis]

    if len(sliceaxis) != 1:
        msg = "Exactly one argument must be given."
        raise ValueError(msg)
    if sliceaxis not in axesdict.keys():
        msg = "Argument name must be one of {}.".format(axesdict.keys())
        raise ValueError(msg)

    info["slice"] = axesdict[sliceaxis]
    axes = tuple(filter(lambda val: val != info["slice"], (0, 1, 2)))
    info["haxis"], info["vaxis"] = axes

    return info


def plot_line(ax, p1, p2, *args, **kwargs):
    """Plot a line between points p1 and p2 on axis ax."""
    ax.plot(*zip(p1, p2), *args, **kwargs)


def plot_box(ax, p1, p2, *args, **kwargs):
    """Plots a cube that spans between p1 and p2 on the given axis.

    Args:
      ax (matplolib axis): matplolib axis object
      p1 (tuple, list, np.ndarray): First cube point
        p1 is of length 3 (xmin, ymin, zmax).
      p2 (tuple, list, np.ndarray): Second cube point
        p2 is of length 3 (xmin, ymin, zmax).
      color (str): matplotlib color string
      linewidth (Real): matplotlib linewidth parameter

    """
    x1, y1, z1 = p1
    x2, y2, z2 = p2

    plot_line(ax, (x1, y1, z1), (x2, y1, z1), *args, **kwargs)
    plot_line(ax, (x1, y2, z1), (x2, y2, z1), *args, **kwargs)
    plot_line(ax, (x1, y1, z2), (x2, y1, z2), *args, **kwargs)
    plot_line(ax, (x1, y2, z2), (x2, y2, z2), *args, **kwargs)

    plot_line(ax, (x1, y1, z1), (x1, y2, z1), *args, **kwargs)
    plot_line(ax, (x2, y1, z1), (x2, y2, z1), *args, **kwargs)
    plot_line(ax, (x1, y1, z2), (x1, y2, z2), *args, **kwargs)
    plot_line(ax, (x2, y1, z2), (x2, y2, z2), *args, **kwargs)

    plot_line(ax, (x1, y1, z1), (x1, y1, z2), *args, **kwargs)
    plot_line(ax, (x2, y1, z1), (x2, y1, z2), *args, **kwargs)
    plot_line(ax, (x1, y2, z1), (x1, y2, z2), *args, **kwargs)
    plot_line(ax, (x2, y2, z1), (x2, y2, z2), *args, **kwargs)


def addcolorbar(ax, imax):
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="3%", pad=0.05)
    cbar = plt.colorbar(imax, cax=cax)
    return ax, cbar



import os
import types
import random
import pytest
import itertools
import numpy as np
import discretisedfield as df


class TestField:
    def setup(self):
        self.meshes = self.create_meshes()
        self.scalar_fs = self.create_scalar_fs()
        self.vector_fs = self.create_vector_fs()
        self.constant_values = [0, -5., np.pi,
                                1e-15, 1.2e12, random.random()]
        self.tuple_values = [(0, 0, 1),
                             (0, -5.1, np.pi),
                             [70, 1e15, 2*np.pi],
                             [5, random.random(), np.pi],
                             np.array([4, -1, 3.7]),
                             np.array([2.1, 0.0, -5*random.random()])]
        self.scalar_pyfuncs = self.create_scalar_pyfuncs()
        self.vector_pyfuncs = self.create_vector_pyfuncs()

    def create_meshes(self):
        p1_list = [(0, 0, 0),
                   (-5e-9, -8e-9, -10e-9),
                   (10, -5, -80)]
        p2_list = [(5e-9, 8e-9, 10e-9),
                   (11e-9, 4e-9, 4e-9),
                   (15, 10, 85)]
        cell_list = [(1e-9, 1e-9, 1e-9),
                     (1e-9, 2e-9, 1e-9),
                     (5, 5, 2.5)]

        meshes = []
        for i in range(len(p1_list)):
            mesh = df.Mesh(p1_list[i], p2_list[i], cell_list[i])
            meshes.append(mesh)
        return meshes

    def create_scalar_fs(self):
        scalar_fs = []
        for mesh in self.meshes:
            f = df.Field(mesh, dim=1)
            scalar_fs.append(f)
        return scalar_fs

    def create_vector_fs(self):
        vector_fs = []
        for mesh in self.meshes:
            f = df.Field(mesh, dim=3)
            vector_fs.append(f)
        return vector_fs

    def create_scalar_pyfuncs(self):
        f = [lambda c: 1,
             lambda c: -2.4,
             lambda c: -6.4e-15,
             lambda c: c[0] + c[1] + c[2] + 1,
             lambda c: (c[0]-1)**2 - c[1]+7 + c[2]*0.1,
             lambda c: np.sin(c[0]) + np.cos(c[1]) - np.sin(2*c[2])]

        return f

    def create_vector_pyfuncs(self):
        f = [lambda c: (1, 2, 0),
             lambda c: (-2.4, 1e-3, 9),
             lambda c: (c[0], c[1], c[2] + 100),
             lambda c: (c[0]+c[2]+10, c[1], c[2]+1),
             lambda c: (c[0]-1, c[1]+70, c[2]*0.1),
             lambda c: (np.sin(c[0]), np.cos(c[1]), -np.sin(2*c[2]))]

        return f

    def test_init(self):
        p1 = (0, -4, 11)
        p2 = (15, 10.1, 16.5)
        d = (1, 0.1, 0.5)
        mesh = df.Mesh(p1, p2, d)
        dim = 2
        name = "test_field"
        value = [1, 2]

        f = df.Field(mesh, dim=2, value=value, name=name)

        assert f.name == name

        assert np.all(f.array[..., 0] == value[0])
        assert np.all(f.array[..., 1] == value[1])

    def test_wrong_init(self):
        mesh = "wrong_mesh_string"
        with pytest.raises(TypeError):
            f = df.Field(mesh, dim=1, name="wrong_field")

    def test_repr(self):
        for f in self.scalar_fs:
            rstr = repr(f)
            assert isinstance(rstr, str)
            assert "dim=1" in rstr
            assert "mesh=" in rstr
            assert "name=" in rstr
            assert rstr.startswith("<")
            assert rstr.endswith(">")

        for f in self.vector_fs:
            rstr = repr(f)
            assert isinstance(rstr, str)
            assert "dim=3" in rstr
            assert "name=" in rstr
            assert rstr.startswith("<")
            assert rstr.endswith(">")

    def test_set_with_constant(self):
        for value, f in itertools.product(self.constant_values,
                                          self.scalar_fs):
            f.value = value

            # Check all values.
            assert np.all(f.value == value)

            # Check with sampling.
            assert np.all(f(f.mesh.random_point()) == value)

    def test_set_with_tuple_list_ndarray(self):
        for value, f in itertools.product(self.tuple_values, self.vector_fs):
            f.value = value
            f.norm = 1

            norm = np.linalg.norm(value)
            for j in range(3):
                c = f.mesh.random_point()
                assert np.all(f.array[..., j] == value[j]/norm)
                assert np.all(f(c)[j] == value[j]/norm)

    def test_norm_is_not_preserved(self):
        for f in self.vector_fs:
            f.value = (0, 3, 0)
            f.norm = 1
            assert np.all(f.norm.array == 1)

            f.value = (0, 2, 0)
            assert np.all(f.norm.array != 1)
            assert np.all(f.norm.array == 2)

    def test_set_with_ndarray(self):
        for f in self.vector_fs:
            value = np.zeros(f.mesh.n + (f.dim,))
            f.value = value

    def test_set_from_callable(self):
        # Test scalar fs.
        for f, pyfun in itertools.product(self.scalar_fs, self.scalar_pyfuncs):
            f.value = pyfun

            for j in range(10):
                c = f.mesh.random_point()
                ev = pyfun(f.mesh.index2point(f.mesh.point2index(c)))
                assert f(c) == ev

        # Test vector fields.
        for f, pyfun in itertools.product(self.vector_fs, self.vector_pyfuncs):
            f.value = pyfun

            for j in range(10):
                c = f.mesh.random_point()
                ev = pyfun(f.mesh.index2point(f.mesh.point2index(c)))
                assert np.all(f(c) == ev)

    def test_set_exception(self):
        for f in self.vector_fs + self.scalar_fs:
            with pytest.raises(TypeError):
                f.value = "string"

    def test_array(self):
        mesh = df.Mesh(p1=(0, 0, 0), p2=(10, 10, 10), cell=(1, 1, 1))
        f = df.Field(mesh, dim=2)

        assert isinstance(f.array, np.ndarray)

    def test_average_scalar_field(self):
        value = -1e-3 + np.pi
        tol = 1e-12
        for f in self.scalar_fs:
            f.value = value
            average = f.average

            assert abs(average[0] - value) < tol

    def test_average_vector_field(self):
        value = np.array([1.1, -2e-3, np.pi])
        tol = 1e-12
        for f in self.vector_fs:
            f.value = value
            average = f.average

            assert abs(average[0] - value[0]) < tol
            assert abs(average[1] - value[1]) < tol
            assert abs(average[2] - value[2]) < tol

    def test_norm_value(self):
        mesh = df.Mesh(p1=(0, 0, 0), p2=(10, 10, 10), cell=(1, 1, 1))
        f = df.Field(mesh, dim=3, value=(2, 2, 2))

        f.norm = 1
        assert isinstance(f.norm.value, int)
        assert np.all(f.norm.value == 1)

        f.array[0, 0, 0, 0] = 3
        assert isinstance(f.norm.value, np.ndarray)

    def test_norm(self):
        mesh = df.Mesh(p1=(0, 0, 0), p2=(10, 10, 10), cell=(1, 1, 1))
        f = df.Field(mesh, dim=3, value=(2, 2, 2))

        f.norm = 1
        assert isinstance(f.norm.value, int)
        assert f.norm.value == 1

        f.array[0, 0, 0, 0] = 3
        assert isinstance(f.norm.value, np.ndarray)

    def test_value(self):
        mesh = df.Mesh(p1=(0, 0, 0), p2=(10, 10, 10), cell=(1, 1, 1))
        f = df.Field(mesh, dim=3)

        f.value = (1, 1, 1)
        assert f.value == (1, 1, 1)

        f.array[0, 0, 0, 0] = 3
        assert isinstance(f.value, np.ndarray)

    def test_normalise(self):
        for f in self.vector_fs:
            for value in self.vector_pyfuncs + self.tuple_values:
                for norm_value in [1, 2.1, 50, 1e-3, np.pi]:
                    f.value = value
                    f.norm = norm_value

                    # Compute norm.
                    norm = 0
                    for j in range(f.dim):
                        norm += f.array[..., j]**2
                    norm = np.sqrt(norm)

                    assert norm.shape == f.mesh.n
                    diff = norm - norm_value

                    assert np.all(abs(norm - norm_value) < 1e-12)

    def test_normalise_scalar_field_exception(self):
        for f, value in itertools.product(self.scalar_fs, self.scalar_pyfuncs):
            for norm_value in [1, 2.1, 50, 1e-3, np.pi]:
                with pytest.raises(ValueError):
                    f.norm = norm_value
                    f.value = value

    def test_norm_zero_field_exception(self):
        for f in self.vector_fs:
            f.value = (0, 0, 0)
            with pytest.raises(ValueError):
                f.norm = 1

    def test_get_vector_field_component(self):
        for f in self.vector_fs:
            f.value = (1, 2, 3)
            assert f.x.dim == 1
            assert np.all(f.x.array == 1)
            assert f.y.dim == 1
            assert np.all(f.y.array == 2)
            assert f.z.dim == 1
            assert np.all(f.z.array == 3)

    def test_get_attr(self):
        for f in self.vector_fs:
            f.value = (1, 2, 3)
            with pytest.raises(AttributeError):
                f.__getattr__("nonexisting_attribute")
            
    def test_dir(self):
        for f in self.vector_fs:
            f.value = (1, 2, 3)
            assert "x" in f.__dir__()
            assert "y" in f.__dir__()
            assert "z" in f.__dir__()
            
    def test_line(self):
        mesh = df.Mesh((0, 0, 0), (10e-9, 10e-9, 10e-9), (1e-9, 1e-9, 1e-9))
        field = df.Field(mesh, value=(1, 2, 3))

        pv = list(field.line(p1=(0, 0, 1e-9), p2=(5e-9, 5e-9, 5e-9)))
        p, v = zip(*pv)

        assert len(p) == 100
        assert len(v) == 100

    def test_plane(self):
        for f in self.vector_fs:
            assert isinstance(f.plane("x"), types.GeneratorType)


    def test_plot_plane(self):
        for f in self.vector_fs:
            f.plot_plane("x")
            f.plot_plane("y")
            f.plot_plane("z")

    def test_write_read_ovf_file_text(self):
        tol = 1e-12
        filename = "test_write_ovf_file_text.omf"
        value = (1e-3 + np.pi, -5, 6)
        for f in self.vector_fs:
            f.value = value
            f.write(filename)

            f_loaded = df.read(filename)

            assert f.mesh.p1 == f_loaded.mesh.p1
            assert f.mesh.p2 == f_loaded.mesh.p2
            assert f.mesh.cell == f_loaded.mesh.cell
            assert f.mesh.cell == f_loaded.mesh.cell
            assert np.all(abs(f.value - f_loaded.value) < tol)

            os.system("rm {}".format(filename))

    def test_write_vector_norm_file(self):
        tol = 1e-12
        mesh = df.Mesh(p1=(0, 0, 0), p2=(10, 10, 10), cell=(5, 5, 5))
        f = df.Field(mesh, dim=1, value=-3.1)

        filename = "test_write_oommf_file_text1.omf"
        f.write(filename)

        f_loaded = df.read(filename)

        assert f.mesh.p1 == f_loaded.mesh.p1
        assert f.mesh.p2 == f_loaded.mesh.p2
        assert f.mesh.cell == f_loaded.mesh.cell
        assert f.mesh.cell == f_loaded.mesh.cell
        assert np.all(abs(f.value - f_loaded.array[..., 0]) < tol)
        assert np.all(abs(0 - f_loaded.array[..., 1]) < tol)
        assert np.all(abs(0 - f_loaded.array[..., 2]) < tol)

        os.system("rm {}".format(filename))

    def test_write_read_ovf_file_binary(self):
        tol = 1e-12
        filename = "test_write_ovf_file_binary.omf"
        value = (1e-3 + np.pi, -5, 6)

        for f in self.vector_fs:
            f.value = value
            f.write(filename, representation="bin8")

            f_loaded = df.read(filename)

            assert f.mesh.p1 == f_loaded.mesh.p1
            assert f.mesh.p2 == f_loaded.mesh.p2
            assert f.mesh.cell == f_loaded.mesh.cell
            assert f.mesh.cell == f_loaded.mesh.cell
            assert np.all(abs(f.value - f_loaded.value) < tol)

        os.system("rm {}".format(filename))

    def test_write_vtk_file(self):
        tol = 1e-12
        filename = "test_write_vtk_file.vtk"
        value = (1e-3 + np.pi, -5, 6)
        for f in self.vector_fs:
            f.value = value
            f.write(filename)

            os.system("rm {}".format(filename))



import pyvtk
import struct
import numpy as np
import joommfutil.typesystem as ts
import discretisedfield as df
import discretisedfield.util as dfu
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable


@ts.typesystem(mesh=ts.Typed(expected_type=df.Mesh),
               dim=ts.Scalar(expected_type=int, unsigned=True, const=True),
               name=ts.Name(const=True))
class Field(dfu.Field):
    """Finite difference field

    This class defines a finite difference field and provides some
    basic operations. The field is defined on a finite difference mesh
    (`discretisedfield.Mesh`).

    Parameters
    ----------
    mesh : discretisedfield.Mesh
        Finite difference rectangular mesh on which the field is defined.
    dim : int, optional
        Dimension of the field value. For instance, if ``dim=3``
        the field is three-dimensional vector field; and for
        ``dim=1`` it is a scalar field.
    value : 0, array_like, callable, optional
        For more details, please refer to the `value` property.
    norm : numbers.Real, callable, optional
        For more details, please refer to the `norm` property.
    name : str, optional
        Field name (the default is "field"). The field name must be a valid
        Python variable name string. More specifically, it must not
        contain spaces, or start with underscore or numeric character.

    Examples
    --------
    Creating a uniform vector field on a nano-sized thin film

    >>> import discretisedfield as df
    >>> p1 = (-50e-9, -25e-9, 0)
    >>> p2 = (50e-9, 25e-9, 5e-9)
    >>> cell = (1e-9, 1e-9, 0.1e-9)
    >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
    >>> value = (0, 0, 1)
    >>> name = "uniform_field"
    >>> field = df.Field(mesh=mesh, dim=3, value=value, name=name)

    .. seealso:: :py:func:`~discretisedfield.Mesh`

    """
    def __init__(self, mesh, dim=3, value=0, norm=None, name="field"):
        self.mesh = mesh
        self.dim = dim
        self.value = value
        self.norm = norm
        self.name = name

    @property
    def value(self):
        """Finite difference field value representation property.

        The getter of this propertry returns a field value
        representation if exists. Otherwise, it returns the
        numpy.ndarray.

        Parameters
        ----------
        value : numbers.Real, array_like, callable
            If the parameter for the setter of this property is 0, all
            values in the field will be set to zero (for vector fields
            ``dim > 1``, all componenets of the vector will be
            0). Input parameter can also be ``numbers.Real`` for
            scalar fields (``dim=1``). In the case of vector fields,
            ``numbers.Real`` is not allowed unless ``value=0``, but an
            array_like data with length equal to the ``dim`` should be
            used. Finally, the value can also be a callable
            (e.g. Python function), which for every coordinate in the
            mesh returns an appropriate value.

        Returns
        -------
        array_like, callable, numbers.Real
            The returned value in the case of vector field can be
            array_like (tuple, list, numpy.ndarray). In the case of
            scalar field, the returned value can be numbers.Real. If
            all components of a vector field are zero, 0 will be
            returned.

        Examples
        --------
        Different ways of setting and getting property `value`.

        >>> import discretisedfield as df
        >>> p1 = (-50e-9, -25e-9, 0)
        >>> p2 = (50e-9, 25e-9, 5e-9)
        >>> cell = (5e-9, 5e-9, 5e-9)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> value = (0, 0, 1)
        >>> name = "uniform_field"
        >>> field = df.Field(mesh=mesh, dim=3, value=value, name=name)
        >>> # Value getter
        >>> field.value
        (0, 0, 1)
        >>> field.value = 0
        >>> field.value
        0
        >>> def value_function(pos):
        ...     x, y, z = pos
        ...     if x <= 0:
        ...         return (0, 0, 1)
        ...     else:
        ...         return (0, 0, -1)
        >>> field.value = value_function
        >>> field((10e-9, 0, 0))
        (0.0, 0.0, -1.0)
        >>> field((-10e-9, 0, 0))
        (0.0, 0.0, 1.0)

        .. note::

           Please note this method is a property and should be called
           as ``field.value``, not ``field.value()``.

        .. seealso:: :py:func:`~discretisedfield.Field.array`

        """
        if np.array_equal(self.array,
                          dfu.as_array(self.mesh, self.dim, self._value)):
            return self._value
        else:
            return self.array

    @value.setter
    def value(self, val):
        self._value = val
        self.array = dfu.as_array(self.mesh, self.dim, val)

    @property
    def array(self):
        """Field value as a numpy array.

        Returns
        -------
        numpy.ndarray
            Field values array.

        Parameters
        ----------
        numpy.ndarray
            The dimensions must be ``(field.mesh.n[0],
            field.mesh.n[1], field.mesh.n[2], dim)``

        Examples
        --------
        Different ways of setting and getting property `value`.

        >>> import discretisedfield as df
        >>> p1 = (0, 0, 0)
        >>> p2 = (1, 1, 1)
        >>> cell = (0.5, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> value = (0, 0, 1)
        >>> name = "uniform_field"
        >>> field = df.Field(mesh=mesh, dim=3, value=value, name=name)
        >>> # array getter
        >>> field.array.shape
        (2, 1, 1, 3)

        .. note::

           Please note this method is a property and should be called
           as ``field.array``, not ``field.array()``.

        .. seealso:: :py:func:`~discretisedfield.Field.value`

        """
        return self._array

    @array.setter
    def array(self, val):
        self._array = val

    @property
    def norm(self):
        current_norm = np.linalg.norm(self.array, axis=self.dim)[..., None]
        if self._norm:
            if np.array_equiv(current_norm, self._norm.array):
                return self._norm

        return Field(self.mesh, dim=1, value=current_norm, name="norm")

    @norm.setter
    def norm(self, val):
        if val:
            if self.dim == 1:
                msg = "Cannot normalise field with dim={}.".format(self.dim)
                raise ValueError(msg)

            if not np.any(self.array):
                msg = "Cannot normalise field with all zero values."
                raise ValueError(msg)

            self._norm = Field(mesh=self.mesh, dim=1, value=val, name="norm")
            self.array /= np.linalg.norm(self.array, axis=self.dim)[..., None]
            self.array *= self._norm.array
        else:
            self._norm = val

    @property
    def average(self):
        """Compute the finite difference field average.

        Returns:
          Finite difference field average.

        """
        return tuple(self.array.mean(axis=(0, 1, 2)))

    def __repr__(self):
        """Representation method."""
        rstr = "<Field(mesh={}, dim={}, name=\"{}\")>"
        return rstr.format(repr(self.mesh), self.dim, self.name)

    def __call__(self, point):
        """Sample the field at `point`.

        Args:
          p (tuple): point coordinate at which the field is sampled

        Returns:
          Field value in cell containing point p

        """
        field_value = self.array[self.mesh.point2index(point)]
        if len(field_value) == 1:
            return field_value
        else:
            return tuple(field_value)

    def __getattr__(self, name):
        if name in list(dfu.axesdict.keys())[:self.dim] and 1 < self.dim <= 3:
            val = self.array[..., dfu.axesdict[name]][..., None]
            fieldname = "{}_{}".format(self.name, name)
            return Field(mesh=self.mesh, dim=1, value=val, name=fieldname)
        else:
            msg = "{} object has no attribute {}."
            raise AttributeError(msg.format(type(self).__name__, name))

    def __dir__(self):
        if 1 < self.dim <= 3:
            extension = list(dfu.axesdict.keys())[:self.dim]
        else:
            extension = []
        return list(self.__dict__.keys()) + extension

    def plane(self, *args, x=None, y=None, z=None, n=None):
        for point in self.mesh.plane(*args, x=x, y=y, z=z, n=n):
            yield point, self.__call__(point)

    def _plot_data(self, *args, x=None, y=None, z=None, n=None, ax=None):
        info = dfu.plane_info(*args, x=x, y=y, z=z)
        data = list(self.plane(*args, x=x, y=y, z=z, n=n))
        ps, vs = list(zip(*data))
        points = list(zip(*ps))
        values = list(zip(*vs))

        if n is None:
            n = (self.mesh.n[info["haxis"]], self.mesh.n[info["vaxis"]])

        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)

        return info, points, values, n, ax

    def imshow(self, *args, x=None, y=None, z=None, n=None, ax=None, **kwargs):
        info, points, values, n, ax = self._plot_data(*args, x=x, y=y, z=z,
                                                      n=n, ax=ax)

        extent = [self.mesh.pmin[info["haxis"]], self.mesh.pmax[info["haxis"]],
                  self.mesh.pmin[info["vaxis"]], self.mesh.pmax[info["vaxis"]]]
        imax = ax.imshow(np.array(values).reshape(n).T, origin="lower",
                         extent=extent, **kwargs)

        return imax

    def quiver(self, *args, x=None, y=None, z=None, n=None, ax=None,
               colour=None, **kwargs):
        info, points, values, n, ax = self._plot_data(*args, x=x, y=y, z=z,
                                                      n=n, ax=ax)

        if not any(values[info["haxis"]] + values[info["vaxis"]]):
            kwargs["scale"] = 1

        if colour is None:
            qvax = ax.quiver(points[info["haxis"]], points[info["vaxis"]],
                             values[info["haxis"]], values[info["vaxis"]],
                             pivot='mid', **kwargs)
        elif colour in dfu.axesdict.keys():
            qvax = ax.quiver(points[info["haxis"]], points[info["vaxis"]],
                             values[info["haxis"]], values[info["vaxis"]],
                             values[dfu.axesdict[colour]],
                             pivot='mid', **kwargs)

        return qvax

    def colorbar(self, ax, colouredplot, cax=None, **kwargs):
        if cax is None:
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="5%", pad=0.1)

        plt.colorbar(colouredplot, cax=cax, **kwargs)

    def plot_plane(self, *args, x=None, y=None, z=None, n=None, ax=None):
        info, points, values, n, ax = self._plot_data(*args, x=x, y=y, z=z,
                                                      n=n, ax=ax)

        if self.dim > 1:
            self.quiver(*args, x=x, y=y, z=z, n=n, ax=ax)
            scfield = getattr(self, list(dfu.axesdict.keys())[info["slice"]])
        else:
            scfield = self

        colouredplot = scfield.imshow(*args, x=x, y=y, z=z, n=n, ax=ax)
        self.colorbar(ax, colouredplot)

        ax.set_xlabel(list(dfu.axesdict.keys())[info["haxis"]])
        ax.set_ylabel(list(dfu.axesdict.keys())[info["vaxis"]])

    def write(self, filename, **kwargs):
        if any([filename.endswith(ext) for ext in [".omf", ".ovf", ".ohf"]]):
            self._writeovf(filename, **kwargs)
        elif filename.endswith(".vtk"):
            self._writevtk(filename)

    def _writevtk(self, filename):
        grid = [pmini + np.linspace(0, li, ni+1) for pmini, li, ni in
                zip(self.mesh.pmin, self.mesh.l, self.mesh.n)]

        structure = pyvtk.RectilinearGrid(*grid)
        vtkdata = pyvtk.VtkData(structure)

        vectors = [self.__call__(i) for i in self.mesh.coordinates]
        vtkdata.cell_data.append(pyvtk.Vectors(vectors, self.name))
        for i, component in enumerate(dfu.axesdict.keys()):
            name = "{}_{}".format(self.name, component)
            vtkdata.cell_data.append(pyvtk.Scalars(list(zip(*vectors))[i],
                                                   name))

        vtkdata.tofile(filename)

    def _writeovf(self, filename, representation="txt"):
        header = ["OOMMF OVF 2.0",
                  "",
                  "Segment count: 1",
                  "",
                  "Begin: Segment",
                  "Begin: Header",
                  "",
                  "Title: Field generated omf file",
                  "Desc: File generated by Field class",
                  "meshunit: m",
                  "meshtype: rectangular",
                  "xbase: {}".format(self.mesh.pmin[0] +
                                     self.mesh.cell[0]/2),
                  "ybase: {}".format(self.mesh.pmin[0] +
                                     self.mesh.cell[1]/2),
                  "zbase: {}".format(self.mesh.pmin[0] +
                                     self.mesh.cell[2]/2),
                  "xnodes: {}".format(self.mesh.n[0]),
                  "ynodes: {}".format(self.mesh.n[1]),
                  "znodes: {}".format(self.mesh.n[2]),
                  "xstepsize: {}".format(self.mesh.cell[0]),
                  "ystepsize: {}".format(self.mesh.cell[1]),
                  "zstepsize: {}".format(self.mesh.cell[2]),
                  "xmin: {}".format(self.mesh.pmin[0]),
                  "ymin: {}".format(self.mesh.pmin[1]),
                  "zmin: {}".format(self.mesh.pmin[2]),
                  "xmax: {}".format(self.mesh.pmax[0]),
                  "ymax: {}".format(self.mesh.pmax[1]),
                  "zmax: {}".format(self.mesh.pmax[2]),
                  "valuedim: {}".format(3),
                  "valuelabels: {0}_x {0}_y {0}_z".format(self.name),
                  "valueunits: A/m A/m A/m",
                  "",
                  "End: Header",
                  ""]

        if representation == "bin4":
            header.append("Begin: Data Binary 4")
            footer = ["End: Data Binary 4",
                      "End: Segment"]
        elif representation == "bin8":
            header.append("Begin: Data Binary 8")
            footer = ["End: Data Binary 8",
                      "End: Segment"]
        elif representation == "txt":
            header.append("Begin: Data Text")
            footer = ["End: Data Text",
                      "End: Segment"]

        f = open(filename, "w")

        # Write header lines to OOMMF file.
        headerstr = "".join(map(lambda line: "# {}\n".format(line), header))
        f.write(headerstr)

        if representation == "bin8":
            # Close the file and reopen with binary write
            # appending to the end of file.
            f.close()
            f = open(filename, "ab")

            # Add the 8 bit binary check value that OOMMF uses
            packarray = [123456789012345.0]
            # Write data lines to OOMMF file.
            for i in self.mesh.indices:
                [packarray.append(vi) for vi in self.array[i]]

            v_binary = struct.pack("d"*len(packarray), *packarray)
            f.write(v_binary)
            f.close()
            f = open(filename, "a")

        else:
            for i in self.mesh.indices:
                if self.dim == 3:
                    v = [vi for vi in self.array[i]]
                elif self.dim == 1:
                    v = [self.array[i][0], 0, 0]
                else:
                    msg = ("Cannot write dim={} field to "
                           "omf file.".format(self.dim))
                    raise TypeError(msg)
                for vi in v:
                    f.write(" " + str(vi))
                f.write("\n")

        # Write footer lines to OOMMF file.
        footerstr = "".join(map(lambda line: "# {}\n".format(line), footer))
        f.write(footerstr)

        f.close()



import abc
import matplotlib.pyplot as plt
from .util import plot_line, plot_box


class Field(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __init__(self): pass  # pragma: no cover

    @property
    @abc.abstractmethod
    def value(self): pass  # pragma: no cover

    @value.setter
    @abc.abstractmethod
    def value(self, val): pass  # pragma: no cover

    @property
    @abc.abstractmethod
    def norm(self): pass  # pragma: no cover

    @norm.setter
    @abc.abstractmethod
    def norm(self, val): pass  # pragma: no cover

    @property
    @abc.abstractmethod
    def average(self): pass  # pragma: no cover

    @abc.abstractmethod
    def __repr__(self): pass  # pragma: no cover

    @abc.abstractmethod
    def __call__(self, point): pass  # pragma: no cover

    @abc.abstractmethod
    def __getattr__(self, name): pass  # pragma: no cover

    @abc.abstractmethod
    def __dir__(self): pass  # pragma: no cover

    @abc.abstractmethod
    def plane(self, *args, x=None, y=None, z=None, n=None):
        pass  # pragma: no cover

    @abc.abstractmethod
    def plot_plane(self, *args, x=None, y=None, z=None, n=None):
        pass  # pragma: no cover

    @abc.abstractmethod
    def write(self, filename, **kwargs): pass  # pragma: no cover

    def line(self, p1, p2, n=100):
        """Slice the field along the line between `p1` and `p2`"""
        for point in self.mesh.line(p1, p2, n=n):
            yield point, self.__call__(point)



import pytest
import pkg_resources
import matplotlib
matplotlib.use("agg")
from .mesh import Mesh
from .field import Field
from .read import read

__version__ = pkg_resources.get_distribution(__name__).version
__dependencies__ = pkg_resources.require(__name__)


def test():
    return pytest.main(["-v", "--pyargs", "discretisedfield"])  # pragma: no cover



import pytest
import numpy as np
import discretisedfield as df


class TestMesh:
    def setup(self):
        self.valid_args = [[(0, 0, 0),
                            (5, 5, 5),
                            [1, 1, 1]],
                           [(0, 0, 0),
                            (5e-9, 5e-9, 5e-9),
                            (1e-9, 1e-9, 1e-9)],
                           [(-1.5e-9, -5e-9, 0),
                            (1.5e-9, -15e-9, -160e-9),
                            (0.15e-9, 0.1e-9, 10e-9)],
                           [(-1.5e-9, -5e-9, -5e-9),
                            np.array((0, 0, 0)),
                            (0.5e-9, 1e-9, 5e-9)],
                           [[0, 5e-6, 0],
                            (-1.5e-6, -5e-6, -5e-6),
                            (0.5e-6, 1e-6, 5e-9)],
                           [(0, 125e-9, 0),
                            (500e-9, 0, -3e-9),
                            (2.5e-9, 2.5e-9, 3e-9)]]

        self.invalid_args = [[(0, 0, 0),
                              (5, 5, 5),
                              (-1, 1, 1)],
                             [("1", 0, 0),
                              (1, 1, 1),
                              (0, 0, 1e-9)],
                             [(-1.5e-9, -5e-9, "a"),
                              (1.5e-9, 15e-9, 16e-9),
                              (5, 1, -1e-9)],
                             [(-1.5e-9, -5e-9, 0),
                              (1.5e-9, 16e-9),
                              (0.1e-9, 0.1e-9, 1e-9)],
                             [(-1.5e-9, -5e-9, 0),
                              (1.5e-9, 15e-9, 1+2j),
                              (5, 1, 1e-9)],
                             ["string",
                              (5, 1, 1e-9),
                              "string"],
                             [(-1.5e-9, -5e-9, 0),
                              (1.5e-9, 15e-9, 16e-9),
                              (2+2j,)]]

    def test_init(self):
        p1 = (0, -4, 16.5)
        p2 = (15, 10.1, 11)
        cell = (1, 0.1, 0.5)
        pbc = 'y'
        name = "test_mesh"
        mesh = df.Mesh(p1=p1, p2=p2, cell=cell, pbc=pbc, name=name)

        assert isinstance(mesh.p1, tuple)
        assert mesh.p1 == p1
        assert isinstance(mesh.p2, tuple)
        assert mesh.p2 == p2
        assert isinstance(mesh.cell, tuple)
        assert mesh.cell == cell
        assert isinstance(mesh.pbc, set)
        assert mesh.pbc == set(pbc)
        assert isinstance(mesh.name, str)
        assert mesh.name == name
        assert isinstance(mesh.pmin, tuple)
        assert mesh.pmin == (0, -4, 11)
        assert isinstance(mesh.pmax, tuple)
        assert mesh.pmax == (15, 10.1, 16.5)

        assert isinstance(mesh.l, tuple)
        assert len(mesh.l) == 3
        assert mesh.l[0] == 15 - 0
        assert mesh.l[1] == 10.1 - (-4)
        assert mesh.l[2] == 16.5 - 11

        assert isinstance(mesh.n, tuple)
        assert len(mesh.n) == 3
        assert mesh.n[0] == (15 - 0) / 1.0
        assert mesh.n[1] == (10.1 - (-4)) / 0.1
        assert mesh.n[2] == (16.5 - 11) / 0.5
        for i in range(3):
            assert isinstance(mesh.n[i], int)

    def test_init_valid_args(self):
        for p1, p2, cell in self.valid_args:
            mesh = df.Mesh(p1=p1, p2=p2, cell=cell)

            assert isinstance(mesh.p1, tuple)
            assert mesh.p1 == tuple(p1)
            assert isinstance(mesh.p2, tuple)
            assert mesh.p2 == tuple(p2)
            assert isinstance(mesh.cell, tuple)
            assert mesh.cell == tuple(cell)
            assert isinstance(mesh.name, str)
            assert mesh.name == "mesh"  # default name value
            assert isinstance(mesh.l, tuple)
            assert isinstance(mesh.n, tuple)
            for i in range(3):
                assert isinstance(mesh.n[i], int)

    def test_init_invalid_args(self):
        for p1, p2, cell in self.invalid_args:
            with pytest.raises(ValueError):
                # Exceptions are raised by descriptors.
                mesh = df.Mesh(p1=p1, p2=p2, cell=cell)

    def test_init_invalid_name(self):
        for name in ["mesh name", "2name", " ", 5, "-name", "+mn"]:
            with pytest.raises((ValueError, TypeError)):
                # Exception is raised by the descriptor (mesh.name).
                mesh = df.Mesh(p1=(0, 0, 0), p2=(10, 10, 10),
                               cell=(1, 1, 1), name=name)

    def test_init_valid_pbc(self):
        for pbc in ["x", "z", "zx", "yxzz", "yz"]:
            mesh = df.Mesh(p1=(0, 0, 0), p2=(10, 10, 10),
                           cell=(1, 1, 1), pbc=pbc)
            assert mesh.pbc == set(pbc)

    def test_init_invalid_pbc(self):
        for pbc in ["abc", "a", "123", 5]:
            with pytest.raises((ValueError, TypeError)):
                # Exception is raised by the descriptor (mesh.pbc).
                mesh = df.Mesh(p1=(0, 0, 0), p2=(10, 10, 10),
                               cell=(1, 1, 1), pbc=pbc)

    def test_zero_domain_edge_length(self):
        args = [[(0, 100e-9, 1e-9),
                 (150e-9, 100e-9, 6e-9),
                 (1e-9, 0.01e-9, 1e-9)],
                [(0, 100e-9, 0),
                 (150e-9, 101e-9, 0),
                 (2e-9, 1e-9, 0.1e-9)],
                [(10e9, 10e3, 0),
                 (0.01e12, 11e3, 5),
                 (1e9, 1e3, 1)]]

        for p1, p2, cell in args:
            with pytest.raises(ValueError) as excinfo:
                mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
            assert "edge length is zero" in str(excinfo)

    def test_domain_not_aggregate_of_cell(self):
        args = [[(0, 100e-9, 1e-9),
                 (150e-9, 120e-9, 6e-9),
                 (4e-9, 1e-9, 1e-9)],
                [(0, 100e-9, 0),
                 (150e-9, 104e-9, 1e-9),
                 (2e-9, 1.5e-9, 0.1e-9)],
                [(10e9, 10e3, 0),
                 (11e9, 11e3, 5),
                 (1e9, 1e3, 1.5)]]

        for p1, p2, cell in args:
            with pytest.raises(ValueError) as excinfo:
                mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
            assert "not an aggregate" in str(excinfo)

    def test_cell_greater_than_domain(self):
        p1 = (0, 0, 0)
        p2 = (1e-9, 1e-9, 1e-9)
        for cell in [(2e-9, 1e-9, 1e-9), (1e-9, 2e-9, 1e-9),
                     (1e-9, 1e-9, 2e-9), (1e-9, 5e-9, 0.1e-9)]:
            with pytest.raises(ValueError) as excinfo:
                mymesh = df.Mesh(p1=p1, p2=p2, cell=cell)
            assert "not an aggregate" in str(excinfo)

    def test_centre(self):
        p1 = (0, 0, 0)
        p2 = (100, 100, 100)
        cell = (1, 1, 1)
        mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        assert mesh.centre == (50, 50, 50)

        p1 = (-18.5, 10, 0)
        p2 = (10, 5, 10)
        cell = (0.1, 0.25, 2)
        mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        assert mesh.centre == (-4.25, 7.5, 5)

        p1 = (500e-9, 125e-9, 3e-9)
        p2 = (0, 0, 0)
        cell = (10e-9, 5e-9, 1e-9)
        mesh = df.Mesh(p1, p2, cell)
        assert mesh.centre == (250e-9, 62.5e-9, 1.5e-9)

    def test_random_point(self):
        p1 = (-18.5, 5, 0)
        p2 = (10, -10, 10e-9)
        cell = (0.1e-9, 0.25, 2e-9)
        mesh = df.Mesh(p1, p2, cell)

        for _ in range(20):
            p = mesh.random_point()
            assert isinstance(p, tuple)
            for i in range(3):
                assert mesh.pmin[i] <= p[i] <= mesh.pmax[i]

    def test_repr_no_pbc(self):
        p1 = (-1, -4, 11)
        p2 = (15, 10.1, 12.5)
        cell = (1, 0.1, 0.5)
        name = "meshname"
        mesh = df.Mesh(p1=p1, p2=p2, cell=cell, name=name)

        rstr = ("Mesh(p1=(-1, -4, 11), p2=(15, 10.1, 12.5), "
                "cell=(1, 0.1, 0.5), pbc=set(), name=\"meshname\")")
        assert repr(mesh) == rstr

    def test_repr_pbc(self):
        p1 = (-1, -4, 11)
        p2 = (15, 10.1, 12.5)
        cell = (1, 0.1, 0.5)
        pbc = "zyx"
        name = "meshname"
        mesh = df.Mesh(p1=p1, p2=p2, cell=cell, pbc=pbc, name=name)

        rstr = ("Mesh(p1=(-1, -4, 11), p2=(15, 10.1, 12.5), "
                "cell=(1, 0.1, 0.5), pbc=\"xyz\", name=\"meshname\")")
        assert repr(mesh) == rstr

    def test_index2point(self):
        p1 = (15, -4, 12.5)
        p2 = (-1, 10.1, 11)
        cell = (1, 0.1, 0.5)
        mesh = df.Mesh(p1, p2, cell)

        assert mesh.index2point((0, 0, 0)) == (-0.5, -3.95, 11.25)
        assert mesh.index2point((5, 10, 1)) == (4.5, -2.95, 11.75)
        assert mesh.index2point((15, 140, 2)) == (14.5, 10.05, 12.25)

        # Correct minimum index
        assert isinstance(mesh.index2point((0, 0, 0)), tuple)
        # Below minimum index
        with pytest.raises(ValueError):
            mesh.index2point((-1, 0, 0))
        with pytest.raises(ValueError):
            mesh.index2point((0, -1, 0))
        with pytest.raises(ValueError):
            mesh.index2point((0, 0, -1))

        # Correct maximum index
        assert isinstance(mesh.index2point((15, 140, 2)), tuple)
        # Above maximum index
        with pytest.raises(ValueError):
            mesh.index2point((16, 0, 0))
        with pytest.raises(ValueError):
            mesh.index2point((0, 141, 0))
        with pytest.raises(ValueError):
            mesh.index2point((0, 0, 3))

    def test_point2index(self):
        p1 = (-10e-9, -5e-9, 10e-9)
        p2 = (10e-9, 5e-9, 0)
        cell = (1e-9, 5e-9, 1e-9)
        mesh = df.Mesh(p1, p2, cell)

        # (0, 0, 0) cell
        assert mesh.point2index((-10e-9, -5e-9, 0)) == (0, 0, 0)
        assert mesh.point2index((-9.5e-9, -2.5e-9, 0.5e-9)) == (0, 0, 0)
        assert mesh.point2index((-9.01e-9, -0.1e-9, 0.9e-9)) == (0, 0, 0)

        # (19, 1, 9) cell
        assert mesh.point2index((10e-9, 5e-9, 10e-9)) == (19, 1, 9)
        assert mesh.point2index((9.5e-9, 2.5e-9, 9.5e-9)) == (19, 1, 9)
        assert mesh.point2index((9.1e-9, 0.1e-9, 9.1e-9)) == (19, 1, 9)

        # vicinity of (0, 0, 0) point
        assert mesh.point2index((1e-16, 1e-16, 0.99e-16)) == (10, 1, 0)
        assert mesh.point2index((-1e-16, -1e-16, 0.01e-16)) == (9, 0, 0)

        p1 = (-10, 5, 0)
        p2 = (10, -5, 10e-9)
        cell = (1, 5, 1e-9)
        mesh = df.Mesh(p1, p2, cell)

        tol = 1e-12  # picometer tolerance
        with pytest.raises(ValueError):
            mesh.point2index((-10-tol, 0, 5))
        with pytest.raises(ValueError):
            mesh.point2index((-5, -5-tol, 5))
        with pytest.raises(ValueError):
            mesh.point2index((-5, 0, -tol))
        with pytest.raises(ValueError):
            mesh.point2index((10+tol, 0, 5))
        with pytest.raises(ValueError):
            mesh.point2index((6, 5+tol, 5))
        with pytest.raises(ValueError):
            mesh.point2index((0, 0, 10e-9+tol))

    def test_index2point_point2index_mutually_inverse(self):
        p1 = (15, -4, 12.5)
        p2 = (-1, 10.1, 11)
        cell = (1, 0.1, 0.5)
        mesh = df.Mesh(p1, p2, cell)

        for i in [(-0.5, -3.95, 11.25), (14.5, 10.05, 12.25)]:
            assert isinstance(mesh.index2point(mesh.point2index(i)), tuple)
            assert mesh.index2point(mesh.point2index(i)) == i

        for i in [(1, 0, 0), (0, 1, 0), (0, 0, 1)]:
            assert isinstance(mesh.point2index(mesh.index2point(i)), tuple)
            assert mesh.point2index(mesh.index2point(i)) == i

    def test_plot(self):
        for arg in self.valid_args:
            p1, p2, cell = arg
            mesh = df.Mesh(p1, p2, cell)
            mesh.plot()

    def test_indices(self):
        p1 = (0, 0, 0)
        p2 = (10, 10, 10)
        cell = (1, 1, 1)
        mesh = df.Mesh(p1, p2, cell)

        counter = 0
        for index in mesh.indices:
            assert isinstance(index, tuple)
            assert len(index) == 3
            for i in range(3):
                assert 0 <= index[i] <= 9
            counter += 1

        assert counter == 1000

    def test_coordinates(self):
        p1 = (0, 0, 0)
        p2 = (10, 10, 10)
        cell = (1, 1, 1)
        mesh = df.Mesh(p1, p2, cell)

        counter = 0
        for coord in mesh.coordinates:
            assert isinstance(coord, tuple)
            assert len(coord) == 3
            for i in range(3):
                assert 0.5 <= coord[i] <= 9.5
            counter += 1

        assert counter == 1000

    def test_line(self):
        p1 = (0, 0, 0)
        p2 = (10, 10, 10)
        cell = (1, 1, 1)
        mesh = df.Mesh(p1, p2, cell)

        tol = 1e-12
        li = mesh.line((0, 0, 0), (10, 10, 10), n=10)
        for point in li:
            assert isinstance(point, tuple)
            assert len(point) == 3

            for coord in point:
                assert 0 <= coord <= 10

        li = list(mesh.line((0, 0, 0), (10, 0, 0), n=30))

        assert len(li) == 30
        assert li[0] == (0, 0, 0)
        assert li[-1] == (10, 0, 0)

    def test_plane(self):
        p1 = (0, 0, 0)
        p2 = (10, 10, 10)
        cell = (1, 1, 1)
        mesh = df.Mesh(p1, p2, cell)

        pi = mesh.plane(z=3, n=(2, 2))
        for point in pi:
            assert isinstance(point, tuple)
            assert len(point) == 3

    def test_script(self):
        for arg in self.valid_args:
            p1, p2, cell = arg
            mesh = df.Mesh(p1, p2, cell)
            with pytest.raises(NotImplementedError):
                script = mesh._script

    def test_issue21(self):
        p1 = (0, 0, 0)
        p2 = (10, 10, 10)
        cell = (1, 1, 1)
        name = "object_name"
        mesh = df.Mesh(p1=p1, p2=p2, cell=cell, name=name)

        assert mesh.p1 == p1
        assert mesh.p2 == p2
        assert mesh.cell == cell
        assert mesh.name == name
        assert mesh.l == (10, 10, 10)
        assert mesh.n == (10, 10, 10)

        # Attempt to change attribute
        with pytest.raises(AttributeError):
            mesh.p2 = (15, 15, 15)
        with pytest.raises(AttributeError):
            mesh.name = "new_object_name"

        assert mesh.p1 == p1
        assert mesh.p2 == p2
        assert mesh.cell == cell
        assert mesh.name == name
        assert mesh.l == (10, 10, 10)
        assert mesh.n == (10, 10, 10)



import random
import itertools
import numpy as np
import matplotlib.pyplot as plt
import joommfutil.typesystem as ts
import discretisedfield.util as dfu
from mpl_toolkits.mplot3d import Axes3D


@ts.typesystem(p1=ts.Vector(size=3, const=True),
               p2=ts.Vector(size=3, const=True),
               cell=ts.Vector(size=3, positive=True, const=True),
               pbc=ts.Subset(sample_set="xyz"),
               name=ts.Name(const=True))
class Mesh:
    """Finite difference rectangular mesh.

    The rectangular mesh domain spans between two points `p1` and
    `p2`. They are defined as ``array_like`` objects of length 3
    (e.g. ``tuple``, ``list``, ``numpy.ndarray``), :math:`p = (p_{x},
    p_{y}, p_{z})`. The domain is then discretised into finite
    difference cells, where dimensions of a single cell are defined
    with a `cell` parameter. Similar to `p1` and `p2`, `cell`
    parameter is an ``array_like`` object :math:`(d_{x}, d_{y},
    d_{z})` of length 3. The parameter `name` is optional and defaults
    to "mesh".

    In order to properly define a mesh, the length of all mesh domain
    edges must be positive, and the mesh domain must be an aggregate
    of discretisation cells.

    Parameters
    ----------
    p1, p2 : (3,) array_like
        Points between which the mesh domain spans :math:`p = (p_{x},
        p_{y}, p_{z})`.
    cell : (3,) array_like
        Discretisation cell size :math:`(d_{x}, d_{y}, d_{z})`.
    pbc : str, optional
        Periodic boundary conditions in x, y, or z direction. Its value
        is a string consisting of one or more of the letters `x`, `y`,
        or ``z, denoting the periodic direction(s).
    name : str, optional
        Mesh name (the default is "mesh"). The mesh name must be a valid
        Python variable name string. More specifically, it must not
        contain spaces, or start with underscore or numeric character.

    Raises
    ------
    ValueError
        If the length of one or more mesh domain edges is zero, or
        mesh domain is not an aggregate of discretisation cells.

    Examples
    --------
    1. Creating a nano-sized thin film mesh

    >>> import discretisedfield as df
    >>> p1 = (-50e-9, -25e-9, 0)
    >>> p2 = (50e-9, 25e-9, 5e-9)
    >>> cell = (1e-9, 1e-9, 0.1e-9)
    >>> name = "mesh_name"
    >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell, pbc="xy", name=name)

    2. An attempt to create a mesh with invalid parameters, so that
    the ``ValueError`` is raised. In this example, the mesh domain is
    not an aggregate of discretisation cells in the :math:`z`
    direction.

    >>> import discretisedfield as df
    >>> p1 = (-25, 3, 0)
    >>> p2 = (25, 6, 1)
    >>> cell = (5, 3, 0.4)
    >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell, name=name)
    Traceback (most recent call last):
        ...
    ValueError: ...

    """
    def __init__(self, p1, p2, cell, pbc={}, name="mesh"):
        self.p1 = tuple(p1)
        self.p2 = tuple(p2)
        self.cell = tuple(cell)
        self.pbc = pbc
        self.name = name

        # Is the length of any mesh domain edges zero?
        for i, li in enumerate(self.l):
            if li == 0:
                msg = "Mesh domain edge length is zero (l[{}]==0).".format(i)
                raise ValueError(msg)

        # Is the mesh domain not an aggregate of discretisation cells?
        tol = 1e-12  # picometre tolerance
        for i, (li, celli) in enumerate(zip(self.l, self.cell)):
            if tol < li % celli < celli - tol:
                msg = ("Mesh domain is not an aggregate of the discretisation "
                       "cell: abs(p2[{0}]-p1[{0}]) % cell[{0}].").format(i)
                raise ValueError(msg)

    @property
    def pmin(self):
        """Mesh point with minimum coordinates.

        The :math:`i`-th component of :math:`p_\\text{min}` is
        computed from points :math:`p_{1}` and :math:`p_{2}` between
        which the mesh domain spans: :math:`p_\\text{min}^{i} =
        \\text{min}(p_{1}^{i}, p_{2}^{i})`.

        Returns
        -------
        tuple (3,)
            Point with minimum mesh coordinates :math:`(p_{x}^\\text{min},
            p_{y}^\\text{min}, p_{z}^\\text{min})`.

        Example
        -------
        Getting the minimum mesh point.

        >>> import discretisedfield as df
        >>> p1 = (-1.1, 2.9, 0)
        >>> p2 = (5, 0, -0.1)
        >>> cell = (0.1, 0.1, 0.005)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> mesh.pmin
        (-1.1, 0, -0.1)

        .. note::

           Please note this method is a property and should be called
           as ``mesh.pmin``, not ``mesh.pmin()``.

        .. seealso:: :py:func:`~discretisedfield.Mesh.pmax`

        """
        return tuple(map(min, zip(self.p1, self.p2)))

    @property
    def pmax(self):
        """Mesh point with maximum coordinates.

        The :math:`i`-th component of :math:`p_\\text{max}` is
        computed from points :math:`p_{1}` and :math:`p_{2}` between
        which the mesh domain spans: :math:`p_\\text{min}^{i} =
        \\text{max}(p_{1}^{i}, p_{2}^{i})`.

        Returns
        -------
        tuple (3,)
            Point with maximum mesh coordinates :math:`(p_{x}^\\text{max},
            p_{y}^\\text{max}, p_{z}^\\text{max})`.

        Example
        -------
        Getting the maximum mesh point.

        >>> import discretisedfield as df
        >>> p1 = (-1.1, 2.9, 0)
        >>> p2 = (5, 0, -0.1)
        >>> cell = (0.1, 0.1, 0.005)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> mesh.pmax
        (5, 2.9, 0)

        .. note::

           Please note this method is a property and should be called
           as ``mesh.pmax``, not ``mesh.pmax()``.

        .. seealso:: :py:func:`~discretisedfield.Mesh.pmin`

        """
        return tuple(map(max, zip(self.p1, self.p2)))

    @property
    def l(self):
        """Mesh domain edge lengths.

        Edge length in any direction :math:`i` is computed from the
        points between which the mesh domain spans :math:`l^{i} =
        |p_{2}^{i} - p_{1}^{i}|`.

        Returns
        -------
        tuple (3,)
            Lengths of mesh domain edges :math:`(l_{x}, l_{y}, l_{z})`.

        Example
        -------
        Getting mesh domain edge lengths.

        >>> import discretisedfield as df
        >>> p1 = (0, 0, -5)
        >>> p2 = (5, 15, 15)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> mesh.l
        (5, 15, 20)

        .. note::

           Please note this method is a property and should be called
           as ``mesh.l``, not ``mesh.l()``.

        """
        return tuple(abs(p1i-p2i) for p1i, p2i in zip(self.p1, self.p2))

    @property
    def n(self):
        """Number of discretisation cells in all directions.

        By dividing the lengths of mesh domain edges with
        discretisation in all directions, the number of cells is
        computed :math:`n^{i} = l^{i}/d^{i}`.

        Returns
        -------
        tuple (3,)
            The number of discretisation cells :math:`(n_{x},
            n_{y}, n_{z})`.

        Example
        -------
        Getting the number of discretisation cells in all three
        directions.

        >>> import discretisedfield as df
        >>> p1 = (0, 5, 0)
        >>> p2 = (5, 15, 1)
        >>> cell = (0.5, 0.1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> mesh.n
        (10, 100, 1)

        .. note::

           Please note this method is a property and should be called
           as ``mesh.n``, not ``mesh.n()``.

        """
        return tuple(int(round(li/di)) for li, di in zip(self.l, self.cell))

    @property
    def centre(self):
        """Mesh domain centre point.

        This point does not necessarily coincide with the
        discretisation cell centre. It is computed as the middle point
        between minimum and maximum coordinate :math:`p_{c}^{i} =
        p_\\text{min}^{i} + 0.5l^{i}`, where :math:`p_\\text{min}^{i}`
        is the :math:`i`-th coordinate of the minimum mesh domain
        point and :math:`l^{i}` is the mesh domain edge length in the
        :math:`i`-th direction.

        Returns
        -------
        tuple (3,)
            Mesh domain centre point :math:`(p_{c}^{x}, p_{c}^{y},
            p_{c}^{z})`.

        Example
        -------
        Getting the mesh centre point.

        >>> import discretisedfield as df
        >>> p1 = (0, 0, 0)
        >>> p2 = (5, 15, 20)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> mesh.centre
        (2.5, 7.5, 10.0)

        .. note::

           Please note this method is a property and should be called
           as ``mesh.centre``, not ``mesh.centre()``.

        """
        return tuple(pmini+0.5*li for pmini, li in zip(self.pmin, self.l))

    @property
    def indices(self):
        """Generator iterating through all mesh cells and yielding their
        indices.

        Yields
        ------
        tuple (3,)
            Mesh cell indices :math:`(i_{x}, i_{y}, i_{z})`.

        Example
        -------
        Getting all mesh cell indices.

        >>> import discretisedfield as df
        >>> p1 = (0, 0, 0)
        >>> p2 = (3, 2, 1)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> tuple(mesh.indices)
        ((0, 0, 0), (1, 0, 0), (2, 0, 0), (0, 1, 0), (1, 1, 0), (2, 1, 0))

        .. note::

           Please note this method is a property and should be called
           as ``mesh.indices``, not ``mesh.indices()``.

        .. seealso:: :py:func:`~discretisedfield.Mesh.coordinates`

        """
        for index in itertools.product(*map(range, reversed(self.n))):
            yield tuple(reversed(index))

    @property
    def coordinates(self):
        """Generator iterating through all mesh cells and yielding their
        coordinates.

        The discretisation cell coordinate corresponds to the cell
        centre point.

        Yields
        ------
        tuple (3,)
            Mesh cell coordinates (`px`, `py`, `pz`).

        Example
        -------
        Getting all mesh cell coordinates.

        >>> import discretisedfield as df
        >>> p1 = (0, 0, 0)
        >>> p2 = (2, 2, 1)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> tuple(mesh.coordinates)
        ((0.5, 0.5, 0.5), (1.5, 0.5, 0.5), (0.5, 1.5, 0.5), (1.5, 1.5, 0.5))

        .. note::

           Please note this method is a property and should be called
           as ``mesh.coordinates``, not ``mesh.coordinates()``.

        .. seealso:: :py:func:`~discretisedfield.Mesh.indices`

        """
        for index in self.indices:
            yield self.index2point(index)

    def __repr__(self):
        """Mesh representation.

        This method returns the string that can be copied in another
        Python script so that exactly the same mesh object could be
        created.

        Returns
        -------
        str
           Mesh representation string.

        Example
        -------
        Getting mesh representation string.

        >>> import discretisedfield as df
        >>> p1 = (0, 0, 0)
        >>> p2 = (2, 2, 1)
        >>> cell = (1, 1, 1)
        >>> pbc = "xy"
        >>> name = "m"
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell, pbc=pbc, name=name)
        >>> repr(mesh)
        'Mesh(p1=(0, 0, 0), p2=(2, 2, 1), cell=(1, 1, 1), pbc="xy", name="m")'

        """
        if self.pbc:
            pbc = "\"{}\"".format("".join(sorted(self.pbc)))
        else:
            pbc = self.pbc
        return ("Mesh(p1={}, p2={}, cell={}, pbc={}, "
                "name=\"{}\")").format(self.p1, self.p2, self.cell,
                                       pbc, self.name)

    def random_point(self):
        """Generate the random point belonging to the mesh.

        Returns
        -------
        tuple (3,)
            Coordinates of a random point inside that belongs to the mesh
            :math:`(x_\\text{rand}, y_\\text{rand}, z_\\text{rand})`.

        Example
        -------
        Getting mesh representation string.

        >>> import discretisedfield as df
        >>> p1 = (0, 0, 0)
        >>> p2 = (200e-9, 200e-9, 1e-9)
        >>> cell = (1e-9, 1e-9, 0.5e-9)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> mesh.random_point()
        (...)

        .. note::

           In the example, ellipsis is used instead of an exact tuple
           because the result differs each time the ``random_point``
           is called.

        """
        return tuple(pmini+random.random()*li
                     for pmini, li in zip(self.pmin, self.l))

    def index2point(self, index):
        """Convert the cell index to its centre's coordinate.

        Parameters
        ----------
        index : (3,) array_like
            The discretisation cell index :math:`(i_{x}, i_{y}, i_{z})`.

        Returns
        -------
            The discretisation cell centre point :math:`(p_{x}, p_{y}, p_{z})`.

        Raises
        ------
            ValueError
                If the discretisation cell index is out of range.

        Example
        -------
        Converting cell index to its coordinate.

        >>> import discretisedfield as df
        >>> p1 = (0, 0, 0)
        >>> p2 = (2, 2, 1)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> mesh.index2point((0, 1, 0))
        (0.5, 1.5, 0.5)

        .. seealso:: :py:func:`~discretisedfield.Mesh.point2index`

        """
        # Does index refer to a cell outside the mesh?
        for i, (indexi, ni) in enumerate(zip(index, self.n)):
            if indexi > ni - 1 or indexi < 0:
                msg = ("index[{}]={} out of range "
                       "[0, n[{}]].").format(i, indexi, ni)
                raise ValueError(msg)
        return tuple(pmini+(indexi+0.5)*celli for pmini, indexi, celli
                     in zip(self.pmin, index, self.cell))

    def point2index(self, p):
        """Convert the mesh coordinate to the cell index which contains it.

        Parameters
        ----------
        p : (3,) array_like
            The mesh point coordinate :math:`(p_{x}, p_{y}, p_{z})`.

        Returns
        -------
            The discretisation cell index :math:`(i_{x}, i_{y}, i_{z})`.

        Raises
        ------
            ValueError
                If `point` is outside the mesh domain.

        Example
        -------
        Converting point coordinate to the cell index.

        >>> import discretisedfield as df
        >>> p1 = (0, 0, 0)
        >>> p2 = (2, 2, 1)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> mesh.point2index((0.2, 1.7, 0.3))
        (0, 1, 0)

        .. seealso:: :py:func:`~discretisedfield.Mesh.index2point`

        """
        self._isoutside(p)

        index = tuple(int(round((pi-pmini)/celli - 0.5))
                      for pi, pmini, celli in zip(p, self.pmin, self.cell))

        # If rounded to the out-of-range values
        return tuple(max(min(ni-1, indexi), 0)
                     for ni, indexi in zip(self.n, index))

    def _isoutside(self, p):
        """Raises ValueError if point is outside the mesh.

        Parameters
        ----------
        p : (3,) array_like
            The mesh point coordinate :math:`(p_{x}, p_{y}, p_{z})`.

        Returns
        -------
            None
                If `point` is inside the mesh.

        Raises
        ------
            ValueError
                If `p` is outside the mesh domain.

        Example
        -------
        Checking if point is outside the mesh.

        >>> import discretisedfield as df
        >>> p1 = (0, 0, 0)
        >>> p2 = (2, 2, 1)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> point1 = (1, 1, 1)
        >>> mesh._isoutside(point1)  # Nothing is returned.
        >>> point2 = (1, 3, 1)
        >>> mesh._isoutside(point2)
        Traceback (most recent call last):
        ...
        ValueError: ...

        """
        for i, (pi, pmini, pmaxi) in enumerate(zip(p, self.pmin, self.pmax)):
            if pi < pmini or pi > pmaxi:
                msg = "Point p[{}]={} outside the mesh.".format(i, pi)
                raise ValueError(msg)

    def line(self, p1, p2, n=100):
        """Line generator.

        Given two points :math:`p_{1}` and :math:`p_{2}`, :math:`n`
        position vectors are generated.

        .. math::

           \\mathbf{r}_{i} = i\\frac{\\mathbf{p}_{2} -
           \\mathbf{p}_{1}}{n-1}

        and this method yields :math:`\\mathbf{r}_{i}` in :math:`n`
        iterations.

        Parameters
        ----------
            p1, p2 : (3,) array_like
                Two points between which the line is generated.
            n : int
                Number of points on the line.

        Yields
        ------
            tuple
                :math:`\\mathbf{r}_{i}`

        Raises
        ------
            ValueError
                If `p` is outside the mesh domain.

        Example
        -------
        >>> import discretisedfield as df
        >>> p1 = (0, 0, 0)
        >>> p2 = (2, 2, 2)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> tuple(mesh.line(p1=(0, 0, 0), p2=(2, 0, 0), n=2))
        ((0.0, 0.0, 0.0), (2.0, 0.0, 0.0))

        """
        self._isoutside(p1)
        self._isoutside(p2)

        p1, p2 = np.array(p1), np.array(p2)
        dl = (p2-p1) / (n-1)
        for i in range(n):
            yield tuple(p1+i*dl)

    def plane(self, *args, x=None, y=None, z=None, n=None):
        info = dfu.plane_info(*args, x=x, y=y, z=z)

        if info["point"] is None:
            info["point"] = self.centre[info["slice"]]
        else:
            test_point = list(self.centre)
            test_point[info["slice"]] = info["point"]
            self._isoutside(test_point)

        if n is None:
            n = (self.n[info["haxis"]], self.n[info["vaxis"]])

        dhaxis, dvaxis = self.l[info["haxis"]]/n[0], self.l[info["vaxis"]]/n[1]
        haxis = np.linspace(self.pmin[info["haxis"]]+dhaxis/2,
                            self.pmax[info["haxis"]]-dhaxis/2, n[0])
        vaxis = np.linspace(self.pmin[info["vaxis"]]+dvaxis/2,
                            self.pmax[info["vaxis"]]-dvaxis/2, n[1])

        for x, y in itertools.product(haxis, vaxis):
            point = 3*[None]
            point[info["slice"]] = info["point"]
            point[info["haxis"]] = x
            point[info["vaxis"]] = y
            yield tuple(point)

    def plot(self):
        """Plots the mesh domain and the discretisation cell.

        This function is called as a display function in Jupyter
        notebook.

        """
        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")
        ax.set_aspect("equal")

        dfu.plot_box(ax, self.pmin, self.pmax, "b-", linewidth=1.5)
        cell_point = tuple(pmini+celli for pmini, celli
                           in zip(self.pmin, self.cell))
        dfu.plot_box(ax, self.pmin, cell_point, "r-", linewidth=1)
        ax.set(xlabel=r"$x$", ylabel=r"$y$", zlabel=r"$z$")

    def _ipython_display_(self):
        """Jupyter notebook mesh representation.

        Mesh domain and discretisation cell are plotted by the
        :py:func:`~discretisedfield.Mesh.plot`.

        """
        self.plot()  # pragma: no cover

    @property
    def _script(self):
        """This abstract method should be implemented by a specific
        calculator.

        Raises
        ------
        NotImplementedError
            If not implemented by a specific calculator.

        """
        raise NotImplementedError



import sys
import struct
import numpy as np
from .mesh import Mesh
from .field import Field


def read(filename, norm=None, name="field"):
    mdatalist = ["xmin", "ymin", "zmin", "xmax", "ymax", "zmax",
                 "xstepsize", "ystepsize", "zstepsize", "valuedim"]
    mdatadict = dict()

    try:
        with open(filename, "r", encoding="utf-8") as ovffile:
            f = ovffile.read()
            lines = f.split("\n")

        mdatalines = filter(lambda s: s.startswith("#"), lines)
        datalines = np.loadtxt(filter(lambda s: not s.startswith("#"), lines))

        for line in mdatalines:
            for mdatum in mdatalist:
                if mdatum in line:
                    mdatadict[mdatum] = float(line.split()[-1])
                    break

    except UnicodeDecodeError:
        with open(filename, "rb") as ovffile:
            f = ovffile.read()
            lines = f.split(b"\n")

        mdatalines = filter(lambda s: s.startswith(bytes("#", "utf-8")), lines)
        datalines = filter(lambda s: not s.startswith(bytes("#", "utf-8")),
                           lines)

        for line in mdatalines:
            for mdatum in mdatalist:
                if bytes(mdatum, "utf-8") in line:
                    mdatadict[mdatum] = float(line.split()[-1])
                    break

        header = b"# Begin: Data Binary "
        data_start = f.find(header)
        header = f[data_start:data_start + len(header) + 1]

        data_start += len(b"# Begin: Data Binary 8\n")
        data_end = f.find(b"# End: Data Binary ")

        if b"4" in header:
            formatstr = "@f"
            checkvalue = 1234567.0
        elif b"8" in header:
            formatstr = "@d"
            checkvalue = 123456789012345.0
            
        if sys.platform == 'win32':
            data_start += 1  # On Windows new line is \n\r.

        listdata = list(struct.iter_unpack(formatstr, f[data_start:data_end]))
        datalines = np.array(listdata)
        
        if datalines[0] != checkvalue:
            raise AssertionError("Something has gone wrong "
                                 "with reading Binary Data")

        datalines = datalines[1:]  # check value removal

    field = _make_field(mdatadict, name)
    r_tuple = tuple(reversed(field.mesh.n)) + (int(mdatadict["valuedim"]),)
    t_tuple = tuple(reversed(range(3))) + (3,)
    field.array = datalines.reshape(r_tuple).transpose(t_tuple)
    field.norm = norm

    return field


def _make_field(mdatadict, name):
    p1 = (mdatadict[key] for key in ["xmin", "ymin", "zmin"])
    p2 = (mdatadict[key] for key in ["xmax", "ymax", "zmax"])
    cell = (mdatadict[key] for key in ["xstepsize", "ystepsize", "zstepsize"])
    dim = int(mdatadict["valuedim"])

    mesh = Mesh(p1=p1, p2=p2, cell=cell)

    return Field(mesh, dim=dim, name=name)



import setuptools

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name='discretisedfield',
    version='0.8.2',
    description=('Python package for creation, reading, analysis, '
                 'and plotting of finite difference fields.'),
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://joommf.github.io',
    author='Marijan Beg, Ryan A. Pepper, Thomas Kluyver, and Hans Fangohr',
    author_email='jupyteroommf@gmail.com',
    packages=setuptools.find_packages(),
    install_requires=['joommfutil',
                      'matplotlib',
                      'pyvtk'],
    classifiers=['Development Status :: 3 - Alpha',
                 'License :: OSI Approved :: BSD License',
                 'Programming Language :: Python :: 3 :: Only',
                 'Operating System :: Unix',
                 'Operating System :: MacOS',
                 'Operating System :: Microsoft :: Windows',
                 'Topic :: Scientific/Engineering :: Physics',
                 'Intended Audience :: Science/Research',
                 'Natural Language :: English']
)
